(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontutil.ml,v 1.17 2001/11/27 03:51:56 malekith Exp $
 *
 * Module interface: Gontutil
 * Description:
 *)

(* Identifier. *)
type id = string

type location = {mutable loc_file: string; 
                 mutable loc_char: int;
                 mutable loc_line: int}

(* Types used by gont compiler. *)
type ty =
    Int                         (* integer of the same size as pointer *)
  | Void                        (* unit type *)
  | Bool                        (* boolean, same size as Int *)
  | Float                       (* C's double *)
  | String
  | Tuple of ty list
  | Func of ft_record
  | TyVar of id
  | Struct of (ty * id) list
  | OptStruct of (ty * id) list
  | Union of (ty * id) list
  | AbstractType
  | NamedType of nt_record      (* <int>List::t, some_struct *)
and typedef = 
   {td_tyvars: id list; 
    td_type: ty;
    td_module: id;              (* "" for current module *)
    td_name: id;
    td_loc: location}
and nt_record = 
   {nt_parms: ty list;
    nt_module: id;
    nt_name: id;
    nt_loc: location;
    mutable nt_tydef: typedef option; (* Filled by Gontbind. *)
    mutable nt_type: ty option        (* Actuall type after substitutions,
                                       * filled by Gonttypes. *)
   }
and ft_record = {ft_ret_type: ty;
                 ft_parms: ty list;
                 mutable ft_tyvars: id list}
;;

class type symbol =
  object
     method get_name : id
     method get_type : ty
     method get_fnc_level : int
     method get_scope_level : int
     method in_closure : bool
     method put_into_closure : unit
  end
                    

type op2 =
    IntPlus
  | IntMinus
  | IntMult
  | IntDiv
  | IntMod
  | FloatPlus
  | FloatMinus
  | FloatMult
  | FloatDiv
  | Assign
  | Eq | Ne | Lt | Gt | Ge | Le
  
type op1 =
    IntNegate
  | FloatNegate
  
type parm_record = {parm_name: id;
                    parm_type: ty;
                    mutable parm_symbol: symbol option}

type raw_exp =
    SimpleBinOp of op2 * exp * exp
  | SimpleUnOp of op1 * exp
  | Ref of ref_record   (* reference to variable *)
  | Call of exp * exp list
  | IntConst of string
  | BoolConst of bool
  | FloatConst of string
  | StringConst of string
  | Lambda of lambda_record
  | CrossModuleRef of xmr_record
and exp = {mutable raw_exp: raw_exp; 
           mutable exp_type: ty option; 
           exp_loc: location}
and ref_record = {ref_name: id;
                  mutable ref_level: int;
                  mutable ref_symbol: symbol option}
and xmr_record = {xmr_module: id;
                  xmr_name: id;
                  mutable xmr_symbol: symbol option}
and raw_stmt =
    Skip
  | Exp of exp
  | Return of exp
  | ReturnVoid
  | VarDef of var_record
  | Block of stmt list
  | Pair of stmt * stmt         (* doesn't introduce new scope *)
  | If of exp * stmt * stmt
  | Loop of stmt
  | BreakIfNot of exp
  | Open of id
and stmt = {raw_stmt: raw_stmt;
            stmt_loc: location}
and var_record = {var_name: id;
                  var_type: ty;
                  mutable var_symbol: symbol option}
and lambda_record = {lambda_ret_type: ty;
                     lambda_parms: parm_record list;
                     lambda_body: stmt;
                     lambda_tyvars: id list}

type fn_decl = {fn_ret_type: ty; 
                fn_name: id; 
                fn_parms: parm_record list;
                fn_body: stmt;
                fn_loc: location;
                fn_tyvars: id list;
                mutable fn_symbol: symbol option;
                fn_export_as: string}

type extern_c_decl = {xc_type: ty; 
                      xc_name: id; 
                      xc_loc: location;
                      mutable xc_symbol: symbol option}

type global_decl = {global_type: ty;
                    global_loc: location;
                    global_name: id;
                    mutable global_symbol: symbol option}

type fn_proto_decl = {fp_type: ty;
                      fp_name: id;
                      fp_loc: location;
                      mutable fp_symbol: symbol option}
                
type decl =
    FuncDecl of fn_decl
  | GlobalDecl of global_decl
  | ExternC of extern_c_decl
  | FuncProto of fn_proto_decl
  | TopLevelOpen of id
  | TypeDef of typedef

exception Syntax_error of location * string
exception Error of string
exception Eof
;;

(* Utils. *)

let xf = Printf.sprintf

let err_msg (loc, msg) =
  let (line, column) = Lexer.get_line_column loc.loc_file loc.loc_char in
  prerr_string (xf "%s:%d:%d: %s" loc.loc_file line column msg);
  prerr_newline ()

let warn (loc, msg) = err_msg (loc, ("warning: " ^ msg))

let err (loc, msg) = raise (Syntax_error(loc, msg))

let debug s = print_string s; print_newline ()

let rec list_apply fnc (c, xxs) =
  match xxs with
      x::xs -> 
        let (c', x') = fnc (c, x) in
        let (c'', xs') = list_apply fnc (c', xs) in
        (c'', x'::xs')
    | [] -> (c, [])

let sequences : (string, int ref) Hashtbl.t = Hashtbl.create 16

let reset () = Hashtbl.clear sequences

let seq_name s =
  let r = 
    (try
       Hashtbl.find sequences s 
     with Not_found -> 
       let n = ref 0 in
       Hashtbl.add sequences s n; n) in
  incr r; s ^ (string_of_int !r)

let rec type_walk f t =
  let fl = List.map f in
  let t' = f t in
  match t' with
      Int | Void | Bool | Float | String | TyVar _ | AbstractType -> t'
    | Tuple l -> Tuple (fl l)
    | Func {ft_ret_type = r; ft_parms = p; ft_tyvars = t} ->
       Func {ft_ret_type = f r; ft_parms = fl p; ft_tyvars = t}
    | Struct l -> Struct (List.map (fun (t, n) -> (f t, n)) l)
    | OptStruct l -> OptStruct (List.map (fun (t, n) -> (f t, n)) l)
    | Union l -> Union (List.map (fun (t, n) -> (f t, n)) l)
    | NamedType ({nt_parms = p;
                  nt_type = None} as r) -> 
      NamedType {r with nt_parms = fl p}
    | NamedType ({nt_parms = p;
                  nt_type = Some tt} as r) -> 
      NamedType {r with nt_parms = fl p; nt_type = Some (f tt)}

let simple_fv t =
  let res : id list ref = ref [] in
  let f t =
    match t with
        TyVar(a) ->
          if not (List.mem a !res) then
            res := a :: !res;
          t
      | NamedType _ | Int | Void | Bool | Float | String | AbstractType
      | Tuple _ | Func _ | Struct _ | OptStruct _ | Union _ -> t in
  ignore(type_walk f t);
  !res

let make_fn_type (r, a) =
  let a' = List.map (fun {parm_type = t} -> t) a in
  let t = {ft_ret_type = r; ft_parms = a'; ft_tyvars = []} in
  let fv = simple_fv (Func t) in
  Func {t with ft_tyvars = fv}
;;   
